@@ -29,18 +29,6 @@
 
 ;;; Code:
 
-;;;;;;;;;;;;;;;;;;;;;
-;; Introduced in 24.3
-
-(when (not (fboundp 'cl-defstruct))
-  (defalias 'cl-defstruct 'defstruct))
-
-(when (not (fboundp 'cl-every))
-  (defalias 'cl-every 'every))
-
-(when (not (fboundp 'cl-subsetp))
-  (defalias 'cl-subsetp 'subsetp))
-
 ;;;;;;;;;;;;;;;;;;;;;
 ;; Introduced in 24.4
 
